/**********************************************************************
		NOTEMIDI - module for serial port MIDI devices

		(C) 2000 by Michael Minn - nycminn@earthlink.net

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   printk messages written to /var/log/messages

***************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/termios.h>
#include <linux/fs.h>
#include <sys/syscall.h>
#include <sys/errno.h>
#include <asm/uaccess.h>

/* #include <bits/fcntl.h>   for open flags */

/* should use an include for this? */
#define O_RDONLY	     00
#define O_WRONLY	     01
#define O_RDWR		     02

/*
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <asm/segment.h>
#include <asm/io.h>
*/

#define NOTEMIDI_MAJOR_NUMBER 93
#define NOTEMIDI_MODULE_NAME  "notemidi"
#define SERIAL_PORT_DEVICE_NAME "/dev/ttyS0"

#ifndef NULL
#define NULL 0
#endif

static struct file *serial_port_file = NULL;
static int current_minor_number = 0;

static ssize_t notemidi_read(struct file *file_pointer, char *data, size_t length, loff_t *offset)
{
	if (length < 1)
		return -EBUSY;
		
	if (serial_port_file && serial_port_file->f_op && serial_port_file->f_op->read)
		return serial_port_file->f_op->read(serial_port_file, data, 1, &serial_port_file->f_pos);
	
	else
	{
		printk(KERN_INFO "No serial port write function\n");
		return -ENODEV;
	}
}

static ssize_t notemidi_write(struct file *file_pointer, const char *data, size_t length, loff_t *offset)
{
	int scan = 0;
	int status = -1;
	char buffer;
	
	if (serial_port_file && serial_port_file->f_op && serial_port_file->f_op->write)
	{
		for (scan = 0; scan < length; ++scan)
		{
			if (current_minor_number > 0) /* MS-124W Burst Byte */
			{
				get_user(buffer, &data[scan]);
				switch(current_minor_number)
				{
					case 2: put_user(0x18, &data[scan]); break;
					case 3: put_user(0x28, &data[scan]); break;
					case 4: put_user(0x48, &data[scan]); break;
					case 5: put_user(0x88, &data[scan]); break;
					default: put_user(0xf8, &data[scan]);
				}
				if ((status = serial_port_file->f_op->write(serial_port_file, &data[scan], 1, &serial_port_file->f_pos)) < 1)
				{
					printk(KERN_INFO "write failure %d\n", status);
					return status;
				}
				put_user(buffer, &data[scan]);
			}
	
			if ((status = serial_port_file->f_op->write(serial_port_file, &data[scan], 1, &serial_port_file->f_pos)) < 1)
			{
				printk(KERN_INFO "write failure %d\n", status);
				return status;
			}
		}
		status = length;
	}
	else
		printk(KERN_INFO "No serial port write function\n");
	
	return status;
}

static int notemidi_open(struct inode *inode, struct file *file_pointer)
{
	printk(KERN_INFO "MS124W being opened by application\n");
	
	if (MOD_IN_USE)
		return -EADDRINUSE;
	
	MOD_INC_USE_COUNT;
	current_minor_number = MINOR(inode->i_dev);
	
	return 0; /* must be 0 or segmentation fault */
}

static int notemidi_release(struct inode *inode, struct file *file_pointer)
{
	printk(KERN_INFO "MS124W being closed by application\n");

	if (MOD_IN_USE)
		MOD_DEC_USE_COUNT;

	return 1;
}
	
/****** Stubs for future development ********

static loff_t notemidi_lseek(struct file *file_pointer, loff_t offset, int direction)
{
	printk("<1> In Lseek\n");
	return 0;
}

static int notemidi_ioctl(struct inode *i_node, struct file *file_pointer, unsigned int number, unsigned long data)
{
	printk(KERN_INFO "In Ioctl\n");
	
	return 0;
}

******************************************/


/*
	loff_t (*llseek) (struct file *, loff_t, int);
	ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
	int (*readdir) (struct file *, void *, filldir_t);
	unsigned int (*poll) (struct file *, struct poll_table_struct *);
	int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
	int (*mmap) (struct file *, struct vm_area_struct *);
	int (*open) (struct inode *, struct file *);
	int (*flush) (struct file *);
	int (*release) (struct inode *, struct file *);
	int (*fsync) (struct file *, struct dentry *);
	int (*fasync) (int, struct file *, int);
	int (*check_media_change) (kdev_t dev);
	int (*revalidate) (kdev_t dev);
	int (*lock) (struct file *, int, struct file_lock *);
*/

static struct file_operations notemidi_fops =
{
	NULL, /* lseek, */
	&notemidi_read,
	&notemidi_write,
	NULL, /* readdir */
	NULL, /* poll */
	NULL, /* ioctl, */
	NULL, /* mmap */
	&notemidi_open,
	NULL, /* flush */
	&notemidi_release,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

int init_module(void)
{
	struct tty_struct *tty;
	struct termios old_termios;
	
	printk(KERN_INFO "Loading MS124W Serial Port MIDI interface module\n");
	register_chrdev(NOTEMIDI_MAJOR_NUMBER, NOTEMIDI_MODULE_NAME, &notemidi_fops);
	
	if (IS_ERR(PTR_ERR(serial_port_file = filp_open(SERIAL_PORT_DEVICE_NAME, O_WRONLY, 0))))
		printk(KERN_INFO "Failure %d opening serial port %s\n", (int) PTR_ERR(serial_port_file), SERIAL_PORT_DEVICE_NAME);
		
	else if ((tty = (struct tty_struct *)serial_port_file->private_data) == NULL)
		printk(KERN_INFO "NULL file private data TTY pointer\n");
	
	else if (tty->driver.set_termios == NULL)
		printk(KERN_INFO "NULL file set_termios function\n");

	else if (tty->termios == NULL)
		printk(KERN_INFO "NULL file termios structure\n");

	else
	{
		old_termios = *tty->termios;
		tty->termios->c_lflag = 0;
		tty->termios->c_lflag = NOFLSH;
		tty->termios->c_iflag = IGNBRK | IGNPAR;
		tty->termios->c_oflag = 0;
		tty->termios->c_cflag =  B38400 | CREAD | CSIZE | CS8 | CLOCAL | HUPCL;
		tty->termios->c_cc[VEOL] = 0; /* '\r'; */
		tty->termios->c_cc[VERASE] = 0;
		tty->termios->c_cc[VKILL] = 0;
		tty->termios->c_cc[VMIN] = 0;
		tty->termios->c_cc[VTIME] = 0;
		(*tty->driver.set_termios)(tty, &old_termios);
	}
		
	return 0;
}

int cleanup_module(void)
{
	if (! IS_ERR(PTR_ERR(serial_port_file)))
		filp_close(serial_port_file, 0);
	
	unregister_chrdev(NOTEMIDI_MAJOR_NUMBER, NOTEMIDI_MODULE_NAME);
	printk(KERN_INFO "Unloading MS124W Serial Port MIDI interface module\n");
	
	return 0;
}

/* COM1 port is /dev/ttyS0 - major 4, minor 64, io address 03F8 */
/* ioctls - TCGETS TCSETS */
/* man ioctl_list */


/* Making a system call from a module:
http://www.linuxdoc.org/LDP/khg/HyperNews/get/khg/158/2.html

#include <sys/syscall.h>

extern long sys_call_table[]; // arch/i386/kernel/entry.S

int (*sys_read)(unsigned int fd, char *buf, int count); // pointer to sys_read. linux/fs/read_write.c

int init_module(void)
{
sys_read = sys_call_table[SYS_read];
 //now you can use sys_read
 sys_read(...);

}
***********/